#!/usr/bin/env python3
import io
import re
from setuptools import setup, find_packages

with io.open('./voc/__init__.py', encoding='utf8') as version_file:
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M)
    if version_match:
        version = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


with io.open('README.rst', encoding='utf8') as readme:
    long_description = readme.read()


setup(
    name='voc',
    version=version,
    description='Tools to convert Python code into Java bytecode.',
    long_description=long_description,
    author='Russell Keith-Magee',
    author_email='russell@keith-magee.com',
    url='http://pybee.org/voc',
    packages=find_packages(exclude=['docs', 'tests']),
    python_requires='>=3.4',
    entry_points={
        'console_scripts': [
            'voc = voc.__main__:main',
            'vod = voc.java.__main__:main',
        ]
    },
    license='New BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
    test_suite='tests',
    package_urls={
        'Funding': 'https://pybee.org/contributing/membership/',
        'Documentation': 'https://voc.readthedocs.io/en/latest/',
        'Tracker': 'https://github.com/pybee/voc/issues',
        'Source': 'https://github.com/pybee/voc',
    },
)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class EnumerateTests(TranspileTestCase):
    def test_enumerate(self):
        self.assertCodeExecution("""
            lst=['a','b','c','d','e']
            print(list(enumerate(lst)))
            lst=['a','b','c','d','e']
            print(list(enumerate(lst,start=-40)))
            lst=['a','b','c','d','e']
            print(list(enumerate(lst,start=46)))
            lst=[('a',4),'b','c',10,'e']
            print(list(enumerate(lst)))
            print(list(enumerate([])))
            print(list(enumerate([], start=10)))
            """)

    def test_enumerate_invalid_start_args(self):
        self.assertCodeExecution("""
            try:
                print(list(enumerate(['a','b','c'], start=None)))
            except TypeError as err:
                print(err)
            try:
                print(list(enumerate(['a','b','c'], start=1.5)))
            except TypeError as err:
                print(err)
            try:
                print(list(enumerate(['a','b','c'], start="start_string")))
            except TypeError as err:
                print(err)
            """)

    def test_enumerate_invalid_iterable(self):
        self.assertCodeExecution("""
            try:
                num=10
                print(list(enumerate(num, start=10)))
            except TypeError as err:
                print(err)
            try:
                print(list(enumerate()))
            except TypeError as err:
                print(err)
            """)


class BuiltinEnumerateFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["enumerate"]

    not_implemented = [
        'test_bytearray',
        'test_bytes',
        'test_dict',
        'test_frozenset',
        'test_list',
        'test_range',
        'test_set',
        'test_str',
        'test_tuple',
    ]



from .. utils import (
TranspileTestCase,
BuiltinFunctionTestCase,
SAMPLE_SUBSTITUTIONS
)


class SetTests(TranspileTestCase):
    pass


class BuiltinSetFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["set"]

    not_implemented = [
        'test_bytes',
        'test_str',
    ]

    not_implemented_versions = {
    }

    is_flakey = [
        'test_dict',
    ]

    substitutions = {
        "{3, 1.2, True}": [
            "{1.2, 3, True}", "{True, 1.2, 3}", "{True, 3, 1.2}", "{3, True, 1.2}", "{1.2, True, 3}"
        ]
    }
    substitutions.update(SAMPLE_SUBSTITUTIONS)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class FrozensetTests(TranspileTestCase):
    pass


class BuiltinFrozensetFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["frozenset"]

    not_implemented = [
        'test_bytes',
        'test_str',
    ]

    is_flakey = [
        'test_dict',
        'test_tuple',
    ]



from ..utils import TranspileTestCase


class DictComprehensionTests(TranspileTestCase):
    def test_syntax(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            d = {v: v**2 for v in x}
            print(len(d))
            print(d[1])
            print(d[2])
            print(d[3])
            print(d[4])
            print(d[5])
            """)

    def test_method(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            d = dict((v, v**2) for v in x)
            print(len(d))
            print(d[1])
            print(d[2])
            print(d[3])
            print(d[4])
            print(d[5])
            """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class RangeTests(TranspileTestCase):
    pass


class BuiltinRangeFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["range"]

    not_implemented = [
        'test_bool',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase, BuiltinTwoargFunctionTestCase


class PowTests(TranspileTestCase):
    def test_int_z(self):
        self.assertCodeExecution("""
            x = 3
            y = 4
            z = 5
            try:
                print(pow(x, y, z))
            except TypeError as err:
                print(err)
        """)

    def test_int_neg_y_pos_z(self):
        self.assertCodeExecution("""
            x = 3
            y = -4
            z = 5
            try:
                print(pow(x, y, z))
            except (TypeError, ValueError) as err:
                print(type(err))
                print(err)
        """)

    def test_int_neg_y_neg_z(self):
        self.assertCodeExecution("""
            x = 3
            y = -4
            z = -5
            try:
                print(pow(x, y, z))
            except (TypeError, ValueError) as err:
                print(type(err))
                print(err)
        """)

    def test_float_x_with_z(self):
        self.assertCodeExecution("""
            x = 3.3
            y = 4
            z = 5
            try:
                print(pow(x, y, z))
            except TypeError as err:
                print(err)
            """)

    def test_float_y_with_z(self):
        self.assertCodeExecution("""
            x = 3
            y = 4.4
            z = 5
            try:
                print(pow(x, y, z))
            except TypeError as err:
                print(err)
            """)

    def test_float(self):
        self.assertCodeExecution("""
            x = 3.3
            y = 4.4
            z = 5.5
            try:
                print(pow(x, y, z))
            except TypeError as err:
                print(err)
        """)

    def test_float_neg_y_with_z(self):
        self.assertCodeExecution("""
            x = 3.3
            y = -4.4
            z = 5.5
            try:
                print(pow(x, y, z))
            except TypeError as err:
                print(err)
        """)


class BuiltinPowFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["pow"]

    not_implemented = [
        'test_bool',
        'test_bytes',
        'test_bytearray',
        'test_class',
        'test_complex',
        'test_dict',
        'test_float',
        'test_frozenset',
        'test_int',
        'test_list',
        'test_None',
        'test_NotImplemented',
        'test_range',
        'test_set',
        'test_slice',
        'test_str',
        'test_tuple',
        'test_obj',
    ]


class BuiltinTwoargPowFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["pow"]

    not_implemented = [
        'test_class_obj',
        'test_obj_class',
    ]

    is_flakey = [
        'test_complex_float',
    ]



from ..utils import TranspileTestCase


class ExtendsTests(TranspileTestCase):
    def test_extends(self):
        self.assertJavaExecution(
            """
            from java.util import HashMap

            class MyHashMap(extends=java.util.HashMap):
                def push_button(self):
                    return "Bing!"

                def put(self, key: java.lang.Object, value: java.lang.Object) -> java.lang.Object:
                    print("I WANT TO SET", key, "TO", value)
                    return super().put(key, value)

            mymap = MyHashMap()
            print(mymap)
            mymap.put("The answer", "42")
            print(mymap)
            print("The answer is", mymap.get("The answer"))
            print("The question is", mymap.get("The question"))
            print("The machine goes " + mymap.push_button())
            print("Done.")
            """,
            """
            {}
            I WANT TO SET The answer TO 42
            {The answer=42}
            The answer is 42
            The question is None
            The machine goes Bing!
            Done.
            """, run_in_function=False)

    def test_no_default_constructor(self):
        self.assertJavaExecution(
            """
            class MyClass(extends=com.example.BaseClass):
                @super({v * 3: int})
                def __init__(self, v):
                    pass

                def peek(self):
                    return self.base_value


            obj = MyClass(42)
            print("POKE:", obj.poke())
            print("PEEK:", obj.peek())
            """,
            java={
                'com/example/BaseClass': """
                package com.example;

                public class BaseClass {
                    public int base_value;

                    public BaseClass(int v) {
                        base_value = v;
                    }

                    public int poke() {
                        return 2 * base_value;
                    }
                }
                """
            },
            out="""
            POKE: 252
            PEEK: 126
            """,
            run_in_function=False)

    def test_protected_access(self):
        self.assertJavaExecution(
            """
            class MyClass(extends=com.example.BaseClass):
                @super({v: int})
                def __init__(self, v):
                    pass

                def peek(self):
                    return self.poke()


            obj = MyClass(42)
            print("PEEK:", obj.peek())
            """,
            java={
                'com/example/BaseClass': """
                package com.example;

                public class BaseClass {
                    public int base_value;

                    public BaseClass(int v) {
                        base_value = v;
                    }

                    protected int poke() {
                        return 2 * base_value;
                    }
                }
                """
            },
            out="""
            PEEK: 84
            """,
            run_in_function=False)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class ChrTests(TranspileTestCase):
    pass


class BuiltinChrFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["chr"]

    not_implemented = [
        'test_complex',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class BytesTests(TranspileTestCase):
    pass


class BuiltinBytesFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["bytes"]



from ..utils import TranspileTestCase


class DescriptorTests(TranspileTestCase):
    def test_getter(self):
        self.assertCodeExecution("""
            class MyObject:
                def getter(self):
                    print("Got attribute")
                    return 1234

                attr = property(getter)

            obj = MyObject()
            print("obj.attr =", obj.attr)

            try:
                obj.attr = 2345
                print("Shouldn't be able to set attribute.")
            except AttributeError:
                print("Couldn't set attribute.")

            try:
                del obj.attr
                print("Shouldn't be able to delete attribute.")
            except AttributeError:
                print("Couldn't delete attribute.")

            print("Done.")
            """)

    def test_getter_and_setter(self):
        self.assertCodeExecution("""
            class MyObject:
                def __init__(self):
                    self._attr = None

                def getter(self):
                    print("Got attribute")
                    return self._attr

                def setter(self, value):
                    print("Setting attribute")
                    self._attr = value * 2
                    print("Attribute set")

                attr = property(getter, setter)

            obj = MyObject()
            print("obj.attr =", obj.attr)
            obj.attr = 2345
            print("obj.attr =", obj.attr)

            try:
                del obj.attr
                print("Shouldn't be able to delete attribute.")
            except AttributeError:
                print("Couldn't delete attribute.")

            print("Done.")
            """)

    def test_getter_and_setter_and_deleter(self):
        self.assertCodeExecution("""
            class MyObject:
                def __init__(self):
                    self._attr = None

                def getter(self):
                    print("Got attribute")
                    return self._attr

                def setter(self, value):
                    print("Setting attribute")
                    self._attr = value * 2
                    print("Attribute set")

                def deleter(self):
                    print("Deleting attribute")
                    self._attr = 42
                    print("Attribute deleted")

                attr = property(getter, setter, deleter)

            obj = MyObject()
            print("obj.attr =", obj.attr)
            obj.attr = 2345
            print("obj.attr =", obj.attr)
            del obj.attr
            try:
                print("obj.attr =", obj.attr)
                print("Shouldn't be able to access attribute.")
            except AttributeError:
                print("Couldn't access attribute.")
            print("Done.")
            """)

    def test_with_decorators(self):
        self.assertCodeExecution("""
            class MyObject:
                def __init__(self):
                    self._attr = None

                @property
                def attr(self):
                    print("Got attribute")
                    return self._attr

                @attr.setter
                def attr(self, value):
                    print("Setting attribute")
                    self._attr = value * 2
                    print("Attribute set")

                @attr.deleter
                def attr(self):
                    print("Deleting attribute")
                    self._attr = 42
                    print("Attribute deleted")

            obj = MyObject()
            print("obj.attr =", obj.attr)
            obj.attr = 2345
            print("obj.attr =", obj.attr)
            del obj.attr
            try:
                print("obj.attr =", obj.attr)
                print("Shouldn't be able to access attribute.")
            except AttributeError:
                print("Couldn't access attribute.")
            print("Done.")
            """)

    def test_with_decorators_misnamed_methods(self):
        self.assertCodeExecution("""
            class MyObject:
                def __init__(self):
                    self._attr = None

                @property
                def attr(self):
                    print("Got attribute")
                    return self._attr

                @attr.setter
                def attr1(self, value):
                    print("Setting attribute")
                    self._attr = value * 2
                    print("Attribute set")

                @attr.deleter
                def attr2(self):
                    print("Deleting attribute")
                    self._attr = 42
                    print("Attribute deleted")

            obj = MyObject()
            print("obj.attr =", obj.attr)
            print("obj.attr1 =", obj.attr1)
            print("obj.attr2 =", obj.attr2)
            obj.attr1 = 2345
            print("obj.attr =", obj.attr)
            print("obj.attr1 =", obj.attr1)
            print("obj.attr2 =", obj.attr2)
            try:
                obj.attr2 = 3456
                print("Shouldn't be able to set value of attr2")
            except AttributeError:
                print("Can't set value of attr2")

            del obj.attr2
            print("obj.attr =", obj.attr)
            print("obj.attr1 =", obj.attr1)
            try:
                print("obj.attr2 =", obj.attr2)
                print("Shouldn't be able to access attribute.")
            except AttributeError:
                print("Couldn't access attribute.")

            print("Done.")
            """)

    def test_with_factory(self):
        self.assertCodeExecution("""
            def make_attr(name, multiplier):
                def getter(self):
                    return getattr(self, '_' + name)

                def setter(self, value):
                    setattr(self, '_' + name, value * multiplier)

                def deleter(self):
                    delattr(self, '_' + name)

                return property(getter, setter, deleter)

            class MyObject:
                doubled = make_attr('doubled', 2)

            MyObject.tripled = make_attr('tripled', 3)

            obj = MyObject()
            obj._doubled = 10
            obj._tripled = 20

            print(obj.doubled)
            print(obj.tripled)
            obj.doubled = 1234
            obj.tripled = 1234
            print(obj.doubled)
            print(obj.tripled)
            del obj.doubled
            del obj.tripled
            try:
                print("obj.doubled =", obj.doubled)
                print("Shouldn't be able to access attribute.")
            except AttributeError:
                print("Couldn't access attribute.")
            try:
                print("obj.tripled =", obj.tripled)
                print("Shouldn't be able to access attribute.")
            except AttributeError:
                print("Couldn't access attribute.")
            """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class SortedTests(TranspileTestCase):
    def test_minimal(self):
        self.assertCodeExecution("""
            samples = [
                ([1, 5, 3, 2, 4, 9, 12], None),
                (["foo", "bar"], None),
                (["foo", "bar"], "invalid"),
                (["one", "two", "three", "four"], len),
                ([(1, 2), (5, 6), (3, 4)], None),
                ([(1, 2), (3, 4), (5, 6, 7)], len),
            ]
            for seq, key in samples:
                try:
                    print('Sample:', seq)
                    print('Sorted:', sorted(seq, key=key))
                    print('Reverse sorted:', sorted(seq, key=key, reverse=True))
                except Exception as e:
                    print(e)
            """, run_in_function=False)


class BuiltinSortedFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["sorted"]

    not_implemented = [
        'test_fozenset',
    ]



# -*- coding: utf-8 -*-
# The Computer Language Benchmarks Game
# http://shootout.alioth.debian.org/
#
# contributed by Antoine Pitrou
# modified by Dominique Wahli and Daniel Nanz

import sys


def make_tree(i, d):

    if d > 0:
        i2 = i + i
        d -= 1
        return (i, make_tree(i2 - 1, d), make_tree(i2, d))
    return (i, None, None)


def check_tree(node):

    (i, l, r) = node
    if l is None:
        return i
    else:
        return i + check_tree(l) - check_tree(r)


def make_check(itde, make=make_tree, check=check_tree):

    i, d = itde
    return check(make(i, d))


def get_argchunks(i, d, chunksize=5000):

    assert chunksize % 2 == 0
    chunk = []
    for k in range(1, i + 1):
        chunk.extend([(k, d), (-k, d)])
        if len(chunk) == chunksize:
            yield chunk
            chunk = []
    if len(chunk) > 0:
        yield chunk


def main(n, min_depth=4):

    max_depth = max(min_depth + 2, n)
    stretch_depth = max_depth + 1

    print('stretch tree of depth %d\t check: %d' % (
          stretch_depth, make_check((0, stretch_depth))))

    long_lived_tree = make_tree(0, max_depth)

    mmd = max_depth + min_depth
    for d in range(min_depth, stretch_depth, 2):
        i = 2 ** (mmd - d)
        cs = 0
        for argchunk in get_argchunks(i, d):
            cs += sum(map(make_check, argchunk))
        print('%d\t trees of depth %d\t check: %d' % (i * 2, d, cs))

    print('long lived tree of depth %d\t check: %d' % (
          max_depth, check_tree(long_lived_tree)))


if __name__ == '__main__':
    for i in range(int(sys.argv[1])):
        main(int(sys.argv[1]))



from ..utils import TranspileTestCase


class IfElifElseTests(TranspileTestCase):
    def test_if(self):
        # Matches if
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('Small x')
            """)

        # No match on if
        self.assertCodeExecution("""
            x = 10
            if x < 5:
                print('Small x')
            """)

        # Test expression statement
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                [1, 2, 3]
                "abc" == "def"
            """)

    def test_if_else(self):
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('Small x')
            else:
                print('Large x')
            """)

        self.assertCodeExecution("""
            x = 10
            if x < 5:
                print('Small x')
            else:
                print('Large x')
            """)

    def test_if_elif_else(self):
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            else:
                print('Large x')
            """)

        self.assertCodeExecution("""
            x = 10
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            else:
                print('Large x')
            """)

        self.assertCodeExecution("""
            x = 100
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            else:
                print('Large x')
            """)

    def test_if_elif_elif_else(self):
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            else:
                print('Huge x')
            """)

        self.assertCodeExecution("""
            x = 10
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            else:
                print('Huge x')
            """)

        self.assertCodeExecution("""
            x = 100
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            else:
                print('Huge x')
            """)

        self.assertCodeExecution("""
            x = 1000
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            else:
                print('Huge x')
            """)

    def test_if_elif_elif(self):
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            """)

        self.assertCodeExecution("""
            x = 10
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            """)

        self.assertCodeExecution("""
            x = 100
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            """)

        self.assertCodeExecution("""
            x = 1000
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            elif x < 500:
                print('Large x')
            """)

    def test_multiple_if(self):
        # Make sure the most recent if block
        # is correctly identified.
        self.assertCodeExecution("""
            x = 1
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            else:
                print('Large x')
            print('Done 1.')

            x = 10
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            else:
                print('Large x')
            print('Done 2.')

            x = 100
            if x < 5:
                print('Small x')
            elif x < 50:
                print('Medium x')
            else:
                print('Large x')

            print('Done 3.')
            """)

    def test_multiple_if_is_comparison(self):
        # Make sure the most recent if block
        # is correctly identified.
        self.assertCodeExecution("""
            x = None
            if x is None:
                print('1: is none')
            else:
                print('1: is not none')
            print('Done 1.')

            if x is not None:
                print('2: is not none')
            else:
                print('2: is none')
            print('Done 2.')
            """)

    def test_end_of_block(self):
        # Ensure that the jump target at the end of a block
        # is correctly identified
        self.assertCodeExecution("""
            x = 100
            if x == 0:
                y = 42
            else:
                y = 37
            z = (x, y)
            print(z)
            print('Done')
            """)

    def test_simple_end_of_function_block_with_return(self):
        # Ensure that if the last instruction in an if block
        # is a return, a GOTO isn't added to the end of the block.
        self.assertCodeExecution("""
            def foo(x):
                print("Testing", x)
                if x == 0:
                    return 42
            y = foo(0)
            print("Result", y)
            y = foo(100)
            print("Result", y)
            print('Done')
            """)

    def test_end_of_function_block_with_return(self):
        # Ensure that if the last instruction in an if/else block
        # is a return, a GOTO isn't added to the end of the block.
        self.assertCodeExecution("""
            def foo(x):
                print("Testing", x)
                if x == 0:
                    return 42
            y = foo(0)
            print("Result", y)
            y = foo(100)
            print("Result", y)
            print('Done')
            """)

    def test_end_of_function_block_with_return_in_else(self):
        # Ensure that if the last instruction in an if/else block
        # is a return, a GOTO isn't added to the end of the block.
        self.assertCodeExecution("""
            def foo(x):
                print("Testing", x)
                if x == 0:
                    return 42
                else:
                    return 37
            y = foo(0)
            print("Result", y)
            y = foo(100)
            print("Result", y)
            y = foo(0)
            print("Result", y)
            print('Done')
            """)

    def test_end_of_block_in_main(self):
        # Ensure that the jump target at the end of a block
        # is correctly identified in a main loop
        self.assertCodeExecution("""
            print('Hello, World')
            if __name__ == '__main__':
                x = 100
                if x == 0:
                    y = 42
                else:
                    y = 37
            """, run_in_function=False)
